﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Text;
using System.Windows.Forms;

using IrrlichtLime;
using IrrlichtLime.Core;
using IrrlichtLime.Video;
using IrrlichtLime.Scene;
using IrrlichtLime.GUI;
using System.Runtime.InteropServices;
using System.Threading.Tasks;


namespace CapstoneProjectCSharp
{
	public partial class WinFormsUserControl : UserControl
	{
        [DllImport("user32.dll")]
        private static extern IntPtr GetForegroundWindow();

        private IntPtr _WindowHandle = IntPtr.Zero;
        private Dimension2Di _WindowSize = new Dimension2Di(0, 0);
        private Vector2Di _WindowPosition = new Vector2Di(0, 0);
        private MainWindow _ParentWindow;
        private Command backgroundCommand;
        private BackgroundWorker backgroundWorker;

        public MainWindow ParentWindow
        {
            get
            {
                return _ParentWindow;
            }
            set
            {
                _ParentWindow = value;
                _WindowSize = new Dimension2Di(this.Width, this.Height);

                if (_ParentWindow != null)
                {
                    _WindowPosition = new Vector2Di((int)_ParentWindow.Left, (int)_ParentWindow.Top);
                    _ParentWindow.LocationChanged += OnMoved;

                    // Save the window handle
                    System.Windows.Window window = System.Windows.Window.GetWindow(_ParentWindow);
                    var wih = new System.Windows.Interop.WindowInteropHelper(window);
                    _WindowHandle = wih.Handle;
                }
            }
        }

		/// <summary>
		/// Simple class to store command data.
		/// We will use it to send commands to the rendering thread
		/// </summary>
		class Command
		{
			public enum Kind
			{
				None,
				Resized,
                Moved
			}

			public Kind Type = Kind.None;
			public object Value = null;

            public int MouseWheelDelta = 0;

			public void SetCommand(Kind type, object value) { Type = type; Value = value; }
			public void Clear() { Type = Kind.None; Value = null; }
		}

		public WinFormsUserControl()
		{
			InitializeComponent();

            this.MouseWheel += OnMouseWheel;

			backgroundCommand = new Command();

			backgroundWorker = new BackgroundWorker();
			backgroundWorker.WorkerSupportsCancellation = true;
			backgroundWorker.DoWork += new DoWorkEventHandler(backgroundWorker_DoWork);
			backgroundWorker.RunWorkerAsync(Handle);

			IsNotifyResizes = true;

            this.Select();
            this.Focus();
		}

		/// <summary>
		/// Indicates that rendering thread is working.
		/// </summary>
		public bool IsRendering { get { return backgroundWorker.IsBusy; } }

		/// <summary>
		/// Indicates that Irrlicht will be notifyed (or not) about resize of client area of this control.
		/// </summary>
		public bool IsNotifyResizes { get; set; }

		/// <summary>
		/// Shuts down the rendering thread.
		/// This method must be called before the controls are disposed.
		/// </summary>
		public void Shutdown()
		{
			if (backgroundWorker.IsBusy)
			{
				backgroundWorker.CancelAsync();

				while (backgroundWorker.IsBusy)
					Application.DoEvents();
			}
		}

		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);

			if (!IsNotifyResizes)
				return;

			lock (backgroundCommand)
			{
				Dimension2Di newSize = new Dimension2Di(this.Width, this.Height);
				if (newSize.Area > 0)
					backgroundCommand.SetCommand(Command.Kind.Resized, newSize);
			}
		}

        private void OnMoved(object sender, EventArgs e)
        {
            if (_ParentWindow != null)
            {
                Vector2Di newPosition = new Vector2Di((int)_ParentWindow.Left, (int)_ParentWindow.Top);
                backgroundCommand.SetCommand(Command.Kind.Moved, newPosition);
            }
        }

        private void OnMouseWheel(object sender, MouseEventArgs e)
        {
            backgroundCommand.MouseWheelDelta = e.Delta;
        }












        delegate Vector2Di MousePointToClient(System.Drawing.Point pnt);

        private Vector2Di GetPointFromClient(System.Drawing.Point pnt)
        {
            if (this.InvokeRequired)
                return (Vector2Di)this.Invoke(new MousePointToClient(GetPointFromClient), pnt);
            else
            {
                pnt = this.PointToClient(Cursor.Position);
                return new Vector2Di(pnt.X + this.Location.X, pnt.Y + this.Location.Y);
            }
        }




        const string contentPath = "../Content/";

        IrrlichtDevice device;
        VideoDriver driver;
        SceneManager smgr;
        GUIEnvironment gui;
        //GPUProgrammingServices gpu;

        Player _Player;
        LightSceneNode light;




        void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker worker = sender as BackgroundWorker;
            IrrlichtCreationParameters p = new IrrlichtCreationParameters();
            //p.DriverType = DriverType.OpenGL;
            p.DriverType = DriverType.Direct3D9;
            p.VSync = true; // Limit FPS to the refresh rate of the monitor
            p.WindowID = (IntPtr)e.Argument;
            p.Fullscreen = true;
            device = IrrlichtDevice.CreateDevice(p);

            if (device == null) return;

            driver = device.VideoDriver;
            smgr = device.SceneManager;
            gui = device.GUIEnvironment;

            // Set scene fog
            //driver->setFog(video::SColor(0,138,125,81), video::EFT_FOG_LINEAR, 250, 1000, .003f, true, false);
            driver.Fog = new Fog(new Color(0, 0, 0, 0), FogType.Linear, 250f, 1000f, 0.003f, true, false);
           



            // Create dummy object
            //AnimatedMesh dummyMesh = smgr.GetMesh(contentPath + "DummyChar.obj");
            //AnimatedMeshSceneNode dummy = smgr.AddAnimatedMeshSceneNode(dummyMesh);

            // Create the lighting
            light = smgr.AddLightSceneNode(null, new Vector3Df(0, 100, 0), new Colorf(1.0f, 1.0f, 1.0f), 600.0f);
            light.ID = 0; // Make it an invalid target for selection.

            LightSceneNode light3 = smgr.AddLightSceneNode(null, new Vector3Df(-100f, 10f, -550f), new Colorf(1.0f, 1.0f, 1.0f), 600.0f);
            light3.ID = 0; // Make it an invalid target for selection.


            // Create & Load the map
            Map map = new Map(device, contentPath);
            //map.LoadMapDataFromFile(@"../Content/MyMap.txt");


            // Create the player & camera
            _Player = new Player(device, contentPath, map);


            
            /*
            // Create dynamic light
            BillboardSceneNode lightBB = smgr.AddBillboardSceneNode();
            lightBB.Position = new Vector3Df(5f, 10f, -550f);
            lightBB.GetMaterial(0).SetTexture(0, driver.GetTexture(contentPath + "particlered.bmp"));
            lightBB.GetMaterial(0).Type = MaterialType.TransparentAddColor;
            lightBB.GetMaterial(0).Lighting = false;
            lightBB.GetMaterial(0).ZWrite = false;
            SceneNodeAnimator fcsna = smgr.CreateFlyCircleAnimator(new Vector3Df(0f, 10f, 0f), 10f, 0.001f);
            lightBB.AddAnimator(fcsna);
            fcsna.Drop();
            LightSceneNode light2 = smgr.AddLightSceneNode(lightBB, new Vector3Df(0f, 10f, 0f), new Colorf(1.0f, 1.0f, 1.0f), 50.0f);
            light2.CastShadows = true;
            light2.LightType = LightType.Point;
            //light2.DebugDataVisible = DebugSceneType.Full;
            light2.ID = 0; // Make it an invalid target for selection.
            lightBB.ID = 0;
            */


            // Create the guards
            List<Guard> guards = new List<Guard>();
            guards.Add(new Guard(device, contentPath));
            guards.Add(new Guard(device, contentPath));
            guards.Add(new Guard(device, contentPath));
            guards.Add(new Guard(device, contentPath));
            guards.Add(new Guard(device, contentPath));

            //foreach (var guard in guards)
            for (int i = 0; i < guards.Count; i++)
            {
                guards[i].LoadWaypathXml(contentPath + "GaurdWaypath" + (i + 1).ToString() + ".xml");
                //guard.LoadWaypathXml(contentPath + "GaurdWaypath1.xml");
                //guards[i].AddWaypoint(new Vector3Df(50), new Vector3Df(0));
                guards[i].SetStartWaypath();
                //guard.SceneNode.DebugDataVisible = DebugSceneType.BBox; //DebugSceneType.Full;
            }


            //SceneNodeAnimator anim = smgr.CreateFlyCircleAnimator(new Vector3Df(0, 15, 0), 30.0f);
            //cam.AddAnimator(anim);
            //anim.Drop();



            // Create the stealable objects
            StealableObject obj1 = new StealableObject(device, contentPath, map, "Vase");
            StealableObject obj2 = new StealableObject(device, contentPath, map, "Vase");
            StealableObject obj3 = new StealableObject(device, contentPath, map, "Vase");
            obj1.SceneNode.Position = new Vector3Df(-202.90651f, 12.46103f, -410.51205f);
            obj2.SceneNode.Position = new Vector3Df(24.41107f, 12.432f, -269.34845f);
            obj3.SceneNode.Position = new Vector3Df(-393.18277f, 12.46103f, -565.05481f);


            // Create the escape point
            EscapePoint escapePoint = new EscapePoint(device, contentPath, map, "Escape Point");

            






            smgr.AddSkyBoxSceneNode(
                contentPath + "irrlicht2_up.jpg",
                contentPath + "irrlicht2_dn.jpg",
                contentPath + "irrlicht2_lf.jpg",
                contentPath + "irrlicht2_rt.jpg",
                contentPath + "irrlicht2_ft.jpg",
                contentPath + "irrlicht2_bk.jpg");

            /*
            gui.AddImage(
                driver.GetTexture(contentPath + "lime_logo_alpha.png"),
                new Vector2Di(30, 0));
            */

            // Game timer
            device.Timer.Start();
            uint timeNow = device.Timer.RealTime;
            uint timeLast = timeNow;
            uint timeDiff = 0;
            uint timesSkipped = 0, timesCounted = 0;

            // Control variables
            Vector2Di mousePos;
            Vector2Di mousePosLast = new Vector2Di(-1);





            // Game loop
            while (device.Run())
            {
                // Game timing
                timeDiff = timeNow - timeLast;

                // Calculate logic around 30 FPS (Frames Per Second)
                if (timeDiff >= 32u) //20u , 33u
                {
                    //Console.WriteLine("Ignored " + timesSkipped.ToString() + " times.");
                    timesSkipped = 0u;

                    timeLast = timeNow;

                    //player.Rotation = new Vector3Df(player.Rotation.X, player.Rotation.Y + 1f, player.Rotation.Z);

                    ////////////////////////////
                    // Player controls
                    Input.UpdateState();
                    ////////////////////////////
                }
                else ++timesSkipped;
                timeNow = device.Timer.RealTime;


                if (!_Player.GameOver)
                {
                    /////////////////// Not Game Over Start ////////////////////


                    /////////////////////////////////
                    // Reset the player walk status
                    _Player.ResetPlayerWalkStatus();

                    IntPtr activeWindow = GetForegroundWindow(); //GetActiveWindow();
                    if (activeWindow == _WindowHandle) //device.WindowFocused) 
                    {
                        if (Input.IsKeyDown(Keys.W))
                        {
                            // Move the player forward
                            _Player.MoveFoward();
                        }
                        if (Input.IsKeyDown(Keys.A))
                        {
                            // Turn the player to the left
                            _Player.TurnLeft();
                        }
                        if (Input.IsKeyDown(Keys.S))
                        {
                            // Move the player backward
                            _Player.MoveBackward();
                        }
                        if (Input.IsKeyDown(Keys.D))
                        {
                            // Turn the player to the right
                            _Player.TurnRight();
                        }
                        if (Input.IsKeyDown(Keys.R))
                        {
                            // Reset Player Position
                            _Player.SceneNode.Position = new Vector3Df(0f, 0f, 0f);
                            continue;
                        }
                        if (Input.IsKeyDown(Keys.Escape))
                        {
                            // End the game
                            Application.Exit();
                            break;
                        }


                        // Misc
                        /*
                        if (Input.WasKeyPressed(Keys.P))
                        {
                            // Create waypoint
                            guards[0].AddWaypoint(_Player.SceneNode.Position, _Player.SceneNode.Rotation);
                            Console.WriteLine("Waypoint Added!");
                        }
                        if (Input.WasKeyPressed(Keys.O))
                        {
                            // Create waypoint
                            guards[0].RemoveLastWaypoint();
                            Console.WriteLine("Last Waypoint Removed!");
                        }
                        if (Input.WasKeyPressed(Keys.L))
                        {
                            // Save the waypath
                            string waypathFile = contentPath + "GaurdWaypath.xml";
                            guards[0].SaveWaypathXml(waypathFile);
                            Console.WriteLine("Waypath saved to: " + waypathFile);
                        }
                        */


                        // Mouse clickary
                        if (Input.IsKeyDown(Keys.LButton))
                        {
                            //Console.WriteLine("Left mouse button clicked!");
                            Vector2Di mouseWinPos = GetPointFromClient(Cursor.Position);

                            if (mouseWinPos.X >= 0 && mouseWinPos.Y >= 0 && mouseWinPos.X <= device.VideoDriver.ViewPort.Width && mouseWinPos.Y <= device.VideoDriver.ViewPort.Height)
                            {
                                mousePos = mouseWinPos;
                                if (mousePosLast.X != -1 && mousePosLast.Y != -1)
                                {
                                    Vector2Di mouseOffset =
                                        new Vector2Di(mousePosLast.X - mousePos.X, mousePosLast.Y - mousePos.Y);
                                    _Player.OrbitCamera(mouseOffset);
                                }
                                mousePosLast = mousePos;
                            }
                            else mousePosLast = new Vector2Di(-1, -1);
                        }
                        else mousePosLast = new Vector2Di(-1, -1);

                        if (Input.WasKeyPressed(Keys.RButton))
                        {
                            Console.WriteLine("Right mouse button clicked!");
                            Vector2Di mouseWinPos = GetPointFromClient(Cursor.Position);

                            if (mouseWinPos.X >= 0 && mouseWinPos.Y >= 0 && mouseWinPos.X <= device.VideoDriver.ViewPort.Width && mouseWinPos.Y <= device.VideoDriver.ViewPort.Height)
                            {
                                //mousePos = mouseWinPos;
                                SceneNode selectedNode = _Player.GetSelectedNodeAtPoint(mouseWinPos);
                                if (selectedNode != null && selectedNode.ID > 0)
                                {
                                    Console.WriteLine("Selected Object: " + selectedNode.Name);
                                    _Player.StealItem(selectedNode);
                                }
                            }
                            ///////////////////////////////////////////////
                        }
                    }


                    // Update the player & camera
                    _Player.Update();

                    // Check if the player is going out the exit...
                    if (_Player.CheckBoundingBoxCollisionWith(escapePoint.SceneNode.BoundingBoxTransformed))
                    {
                        _Player.GameOver = true;
                        _Player.PlayerExited = true;
                    }

                    // Do guard logic
                    foreach (var guard in guards)
                    {
                        guard.Update();

                        // check collision
                        //if (_Player.CheckBoundingBoxCollisionWith(guard.SceneNode.BoundingBoxTransformed))
                        if (_Player.CheckBoundingBoxCollisionWith(guard.LineOfSite))
                        {
                            guard.AllowMovement = false;
                            _Player.WasSpotted = true;
                            Console.WriteLine("Collision!");
                        }
                    }
                    /////////////////// Not Game Over End ////////////////////
                }


                // Start Rendering the scene
                driver.BeginScene(false);

                smgr.DrawAll();
                gui.DrawAll();

                // Draw player interface
                _Player.UpdateGui();

                // Done Rendering the scene
                driver.EndScene();


                // Check for cancellation
                if (worker.CancellationPending)
                    device.Close();


                // Check for new command
                lock (backgroundCommand)
                {
                    switch (backgroundCommand.Type)
                    {
                        case Command.Kind.Resized:
                            _WindowSize = backgroundCommand.Value as Dimension2Di;
                            driver.ResizeNotify(_WindowSize);
                            backgroundCommand.Clear();
                            break;
                        case Command.Kind.Moved:
                            _WindowPosition = backgroundCommand.Value as Vector2Di;
                            backgroundCommand.Clear();
                            break;
                    }

                    if (backgroundCommand.MouseWheelDelta != 0)
                        _Player.ZoomCamera(ref backgroundCommand.MouseWheelDelta);
                }
                ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            }

            // Game is over!
            //System.Windows.Forms.MessageBox.Show("Game Over!");

            // Stop the engine & Clean-up
            device.Timer.Stop();
            device.Drop();
            Application.Exit();
        }
        ////////////////////////////////////////////////////////////////////////////////////////////////
	}
}
